#!/usr/bin/python
# -*- coding: utf-8 -*-
import _tripack
import _srfpack
import numpy as np

try: range = xrange
except: pass

__version__ = "1.1"

_ier_codes = {0:  "no errors were encountered.",
              -1: "N < 3 on input.",
              -2: "the first three nodes are collinear.",
              -3: "duplicate nodes were encountered.",
              -4: "an error flag was returned by a call to SWAP in ADDNOD.\n \
                   This is an internal error and should be reported to the programmer.",
              'L':"nodes L and M coincide for some M > L.\n \
                   The linked list represents a triangulation of nodes 1 to M-1 in this case.",
              1: "NCC, N, NROW, or an LCC entry is outside its valid range on input.",
              2: "the triangulation data structure (LIST,LPTR,LEND) is invalid.",
              'K': 'NPTS(K) is not a valid index in the range 1 to N.'}


class Triangulation(object):
    """
    Define a Delaunay triangulation for given Cartesian mesh points (x, y)
    where x and y vectors are 1D numpy arrays of equal length.

    Algorithm
    ---------
     R. J. Renka (1996), Algorithm 751; TRIPACK: a constrained two-
     dimensional Delaunay triangulation package,
     ACM Trans. Math. Softw., 22(1), pp 1–8,
     doi:10.1145/225545.225546.

    Parameters
    ----------
     x : 1D array of Cartesian x coordinates
     y : 1D array of Cartesian y coordinates

    Attributes
    ----------
     x : array of floats, shape (n,)
        stored Cartesian x coordinates from input
     y : array of floats, shape (n,)
        stored Cartesian y coordinates from input
     simplices : array of ints, shape (nsimplex, 3)
        indices of the points forming the simplices in the triangulation
        points are ordered anticlockwise
     lst : array of ints, shape (6n-12,)
        nodal indices with lptr and lend, define the triangulation as a set of N
        adjacency lists; counterclockwise-ordered sequences of neighboring nodes
        such that the first and last neighbors of a boundary node are boundary
        nodes (the first neighbor of an interior node is arbitrary).  In order to
        distinguish between interior and boundary nodes, the last neighbor of
        each boundary node is represented by the negative of its index.
        The indices are 1-based (as in Fortran), not zero based (as in python).
     lptr : array of ints, shape (6n-12),)
        set of pointers in one-to-one correspondence with the elements of lst.
        lst(lptr(i)) indexes the node which follows lst(i) in cyclical
        counterclockwise order (the first neighbor follows the last neighbor).
        The indices are 1-based (as in Fortran), not zero based (as in python).
     lend : array of ints, shape (n,)
        N pointers to adjacency lists.
        lend(k) points to the last neighbor of node K. 
        lst(lend(K)) < 0 if and only if K is a boundary node.
        The indices are 1-based (as in Fortran), not zero based (as in python).

    Notes
    -----
     Provided the nodes are randomly ordered, the algorithm
     has expected time complexity O(N*log(N)) for most nodal
     distributions.  Note, however, that the complexity may be
     as high as O(N**2) if, for example, the nodes are ordered
     on increasing latitude.
    """
    def __init__(self, x, y):

        if len(x.shape) != 1 or len(y.shape) != 1:
            raise ValueError('x and y must be 1D')

        if x.size != y.size:
            raise ValueError('x and y must have same length')

        # Triangulation
        lst, lptr, lend, ierr = _tripack.trmesh(x,y)

        if ierr != 0:
            raise ValueError('ierr={} in trmesh\n{}'.format(ierr, _ier_codes[ierr]))

        self.x = x
        self.y = y
        self.points = np.column_stack([x,y])
        self.npoints = x.size

        self.lst = lst
        self.lptr = lptr
        self.lend = lend

        # Convert a triangulation to a triangle list form (human readable)
        # Uses an optimised version of trlist that returns triangles
        # without neighbours or arc indices
        nt, ltri, ierr = _tripack.trlist2(lst, lptr, lend)

        if ierr != 0:
            raise ValueError('ierr={} in trlist2\n{}'.format(ierr, _ier_codes[ierr]))

        # extract triangle list and convert to zero-based ordering
        self.simplices = ltri.T[:nt] - 1


    def neighbour_simplices(self):
        """
        Get indices of neighbour simplices for each simplex.
        The kth neighbour is opposite to the kth vertex.
        For simplices at the boundary, -1 denotes no neighbour.
        """
        nt, ltri, lct, ierr = _tripack.trlist(self.lst, self.lptr, self.lend, nrow=6)
        if ierr != 0:
            raise ValueError('ierr={} in trlist\n{}'.format(ierr, _ier_codes[ierr]))
        return ltri.T[:nt,3:] - 1


    def neighbour_and_arc_simplices(self):
        """
        Get indices of neighbour simplices for each simplex and arc indices.
        Identical to get_neighbour_simplices() but also returns an array
        of indices that reside on boundary hull, -1 denotes no neighbour.
        """
        nt, ltri, lct, ierr = _tripack.trlist(self.lst, self.lptr, self.lend, nrow=9)
        if ierr != 0:
            raise ValueError('ierr={} in trlist\n{}'.format(ierr, _ier_codes[ierr]))
        ltri = ltri.T[:nt] - 1
        return ltri[:,3:6], ltri[:,6:]


    def nearest_neighbour(self, xi, yi):
        """
        Get the index of the nearest neighbour to a point (xi,yi)
        and return the squared distance between (xi,yi) and
        its nearest neighbour.

        Notes
        -----
         Faster searches can be obtained using a KDTree.
         Store all x and y coordinates in a KDTree, then query
         a set of points to find their nearest neighbours.
        """
        # i is the node at which we start the search
        # the closest x coordinate is a good place
        i = ((self.x - xi)**2).argmin() + 1
        idx, d = _tripack.nearnd(xi, yi, i, self.x, self.y, self.lst, self.lptr, self.lend)
        return idx - 1, d


    def gradient(self, f, nit=3, tol=1e-3, guarantee_convergence=False):
        """
        Return the gradient of an n-dimensional array.

        The method consists of minimizing a quadratic functional Q(G) over
        gradient vectors (in x and y directions), where Q is an approximation
        to the linearized curvature over the triangulation of a C-1 bivariate
        function F(x,y) which interpolates the nodal values and gradients.

        Parameters
        ----------
         f : array of floats, shape (n,)
            field over which to evaluate the gradient
         nit: int (default: 3)
            number of iterations to reach a convergence tolerance, tol
            nit >= 1
         tol: float (default: 1e-3)
            maximum change in gradient between iterations.
            convergence is reached when this condition is met.

        Returns
        -------
         dfdx : array of floats, shape (n,)
            derivative of f in the x direction
         dfdy : array of floats, shape (n,)
            derivative of f in the y direction

        Notes
        -----
         For SIGMA = 0, optimal efficiency was achieved in testing with
         tol = 0, and nit = 3 or 4.

         The restriction of F to an arc of the triangulation is taken to be
         the Hermite interpolatory tension spline defined by the data values
         and tangential gradient components at the endpoints of the arc, and
         Q is the sum over the triangulation arcs, excluding interior
         constraint arcs, of the linearized curvatures of F along the arcs --
         the integrals over the arcs of D2F(T)**2, where D2F(T) is the second
         derivative of F with respect to distance T along the arc.
        """
        if f.size != self.npoints:
            raise ValueError('f should be the same size as mesh')

        gradient = np.zeros((2,self.npoints), order='F', dtype=np.float32)
        sigma = 0
        iflgs = 0

        ierr = 1
        while ierr == 1:
            ierr = _srfpack.gradg(self.x, self.y, f, self.lst, self.lptr, self.lend,\
                                  iflgs, sigma, gradient, nit=nit, dgmax=tol)
            if not guarantee_convergence:
                break

        if ierr < 0:
            raise ValueError('ierr={} in gradg\n{}'.format(ierr, _ier_codes[ierr]))

        return gradient[0], gradient[1]


    def gradient_local(self, f, index):
        """
        Return the gradient at a specified node.

        This routine employs a local method, in which values depend only on nearby
        data points, to compute an estimated gradient at a node.

        gradient_local() is more efficient than gradient() only if it is unnecessary
        to compute gradients at all of the nodes. Both routines have similar accuracy.

        Parameters
        ----------
        """
        if f.size != self.npoints:
            raise ValueError('f should be the same size as mesh')


        gradX, gradY, l = _srfpack.gradl(index + 1, self.x, self.y, f,\
                                         self.lst, self.lptr, self.lend)

        return gradX, gradY


    def interpolate_nearest(self, xi, yi, zdata):
        """
        Nearest-neighbour interpolation.
        Calls nearnd to find the index of the closest neighbours to xi,yi

        Parameters
        ----------
         xi : float / array of floats, shape (l,)
            x coordinates on the Cartesian plane
         yi : float / array of floats, shape (l,)
            y coordinates on the Cartesian plane

        Returns 
        -------
         zi : float / array of floats, shape (l,)
            nearest-neighbour interpolated value(s) of (xi,yi)
        """
        if zdata.size != self.npoints:
            raise ValueError('zdata should be same size as mesh')

        ist = np.ones_like(xi)
        ist, dist = _tripack.nearnds(xi, yi, ist, self.x, self.y,
                                     self.lst, self.lptr, self.lend)
        return zdata[ist - 1]


    def interpolate_linear(self, xi, yi, zdata):
        """
        Piecewise linear interpolation/extrapolation to arbitrary point(s).
        The method is fast, but has only C^0 continuity.

        Parameters
        ----------
         xi : float / array of floats, shape (l,)
            x coordinates on the Cartesian plane
         yi : float / array of floats, shape (l,)
            y coordinates on the Cartesian plane
         zdata : array of floats, shape (n,)
            value at each point in the triangulation
            must be the same size of the mesh

        Returns
        -------
         zi : float / array of floats, shape (l,)
            interpolated value(s) of (xi,yi)
        """

        if zdata.size != self.npoints:
            raise ValueError('zdata should be same size as mesh')

        if np.array(xi).size > 1:
            if xi.size != yi.size:
                raise ValueError('xi and yi must have same length')

            n = xi.size
            zi = np.empty(n)
            ist = 1
            
            # iterate
            for i in range(0,n):
                zi[i], t = _srfpack.intrc0(xi[i], yi[i], self.x, self.y, zdata,\
                                           self.lst, self.lptr, self.lend, ist)
        else:
            ist = 1
            zi, t = _srfpack.intrc0(xi, yi, self.x, self.y, zdata,\
                                    self.lst, self.lptr, self.lend, ist)

        return zi


    def interpolate_cubic(self, xi, yi, zdata, gradz=None, derivatives=False):
        """
        Cubic spline interpolation/extrapolation to arbirary point(s).
        This method has C^1 continuity.

        Parameters
        ----------
         xi : float / array of floats, shape (l,)
            x coordinates on the Cartesian plane
         yi : float / array of floats, shape (l,)
            y coordinates on the Cartesian plane
         zdata : array of floats, shape (n,)
            value at each point in the triangulation
            must be the same size of the mesh
         gradz (optional) : array of floats, shape (2,n)
            derivative at each point in the triangulation in the
            x-direction (first row), y-direction (second row)
            if not supplied it is evaluated using the gradg routine
         derivatives (optional) : bool (default: False)
            optionally returns the first derivatives at point(s) (xi,yi)

        Returns
        -------
         zi : float / array of floats, shape (l,)
            interpolated value(s) of (xi,yi)
         dzx, dzy (optional) : float, array of floats, shape(l,)
            first partial derivatives in x and y direction at (xi,yi)
        """

        if zdata.size != self.npoints:
            raise ValueError('zdata should be same size as mesh')

        iflgs = 0
        dflag = 1
        sigma = 0.0
        ist = 1

        xi = np.array(xi)
        yi = np.array(yi)

        if gradz == None:
            gradz = np.zeros((2,self.npoints), order='F', dtype=np.float32)
            ierr = _srfpack.gradg(self.x, self.y, zdata, self.lst, self.lptr, self.lend,\
                                  iflgs, sigma, gradz, nit=3, dgmax=1e-3)
            if ierr < 0:
                raise ValueError('ierr={} in gradg\n{}'.format(ierr, _ier_codes[ierr]))


        if xi.size > 1:
            if xi.size != yi.size:
                raise ValueError('xi and yi must have same length')

            n = xi.size
            zi  = np.empty(n)
            dzx = np.empty(n)
            dzy = np.empty(n)
            
            # iterate
            for i in range(0,n):
                zi[i], dzx[i], dzy[i], ierr = _srfpack.intrc1(xi[i], yi[i], self.x, self.y, zdata,\
                                   self.lst, self.lptr, self.lend, iflgs, sigma, gradz, dflag, ist)
        else:
            zi, dzx, dzy, ierr = _srfpack.intrc1(xi, yi, self.x, self.y, zdata,\
                      self.lst, self.lptr, self.lend, iflgs, sigma, gradz, dflag, ist)

        if ierr < 0:
            raise ValueError('ierr={} in gradg\n{}'.format(ierr, _ier_codes[ierr]))

        if derivatives:
            return zi, (dzx, dzy)
        else:
            return zi


    def smoothing(self, f, w, sm, smtol, gstol):
        """
        Smooths a surface f by choosing nodal function values and gradients to
        minimize the linearized curvature of F subject to a bound on the
        deviation from the data values. This is more appropriate than interpolation
        when significant errors are present in the data.

        Parameters
        ----------
         f : array of floats, shape (n,)
            field to apply smoothing on
         w : array of floats, shape (n,)
            weights associated with data value in f
            w[i] = 1/sigma_f^2 is a good rule of thumb.
         sm : float
            positive parameter specifying an upper bound on Q2(f).
            generally n-sqrt(2n) <= sm <= n+sqrt(2n)
         smtol : float
            specifies relative error in satisfying the constraint
            sm(1-smtol) <= Q2 <= sm(1+smtol) between 0 and 1.
         gstol : float
            tolerance for convergence.
            gstol = 0.05*mean(sigma_f)^2 is a good rule of thumb.

        Returns
        -------
         f_smooth : array of floats, shape (n,)
            smoothed version of f
         (dfdx, dfdy) : tuple of floats, tuple of 2 shape (n,) arrays
            first derivatives of f_smooth in the x and y directions


        """
        if f.size != self.npoints or f.size != w.size:
            raise ValueError('f and w should be the same size as mesh')

        sigma = 0
        iflgs = 0

        f_smooth, df, ierr = _srfpack.smsurf(self.x, self.y, f, self.lst, self.lptr, self.lend,\
                                             iflgs, sigma, w, sm, smtol, gstol)

        if ierr < 0:
            raise ValueError('ierr={} in gradg\n{}'.format(ierr, _ier_codes[ierr]))
        if ierr == 1:
            raise RuntimeWarning("No errors were encountered but the constraint is not active --\n\
                  F, FX, and FY are the values and partials of a linear function \
                  which minimizes Q2(F), and Q1 = 0.")

        return f_smooth, (df[0], df[1])


    def convex_hull(self):
        """
        Find the Convex Hull of the set of x,y points.

        Returns
        -------
         bnodes : array of ints
            indices corresponding to points on the convex hull
        """
        bnodes, nb, na, nt = _tripack.bnodes(self.lst, self.lptr, self.lend, self.npoints)
        return bnodes[:nb] - 1


    def areas(self):
        """
        Compute the area each triangle within the triangulation of points.

        Returns
        -------
         area : array of floats, shape (nt,)
            area of each triangle in self.simplices where nt
            is the number of triangles.

        """
        v1 = self.points[self.simplices[:,1]] - self.points[self.simplices[:,0]]
        v2 = self.points[self.simplices[:,2]] - self.points[self.simplices[:,1]]

        area = 0.5*(v1[:,0]*v2[:,1] - v1[:,1]*v2[:,0])
        return area


import numpy as np
import stripy
from time import clock


def time_routine(routine, *args):
	t = clock()
	routine(*args)
	str_fmt = "- {:25} completed in  {:1.6f} seconds"
	print(str_fmt.format(str(routine.__name__), clock()-t))

"""
Test Triangulation routines
"""
print("---\nTriangulation routines\n---")

# Create some (semi) random points
np.random.seed(0)
x = np.random.random(10000)
y = np.random.random(10000)

z = np.hypot(x,y)

xi = np.random.random(100)
yi = np.random.random(100)

# triangulation
tri = stripy.Triangulation(x, y)

time_routine(tri.__init__, x, y)
time_routine(tri.areas)
time_routine(tri.convex_hull)
time_routine(tri.gradient, z)
time_routine(tri.gradient_local, z, 0)
time_routine(tri.interpolate_nearest, xi, yi, z)
time_routine(tri.interpolate_linear, xi, yi, z)
time_routine(tri.interpolate_cubic, xi, yi, z)
time_routine(tri.nearest_neighbour, 0.5, 0.5)
time_routine(tri.neighbour_simplices)
time_routine(tri.neighbour_and_arc_simplices)


"""
Test  sTriangulation routines
"""
print("---\nsTriangulation routines\n---")

# Create some (semi) random points
np.random.seed(0)
lons = 2.*np.pi*np.random.random(10000)
lats = np.arccos(2.*np.random.random(10000) - 1.) - np.pi/2
Z = np.exp(-lons**2 - lats**2)

lon = 2.*np.pi*np.random.random(100)
lat = np.arccos(2.*np.random.random(100) - 1.) - np.pi/2

stri = stripy.sTriangulation(lons, lats)

time_routine(stri.__init__, lons, lats)
time_routine(stri.areas)
time_routine(stri.gradient, Z)
time_routine(stri.interpolate_nearest, lon, lat, Z)
time_routine(stri.interpolate_linear, lon, lat, Z)
time_routine(stri.interpolate_cubic, lon, lat, Z)
time_routine(stri.transform_to_spherical, Z/stri.x, Z/stri.y, Z/stri.z)
time_routine(stri.tri_area, lon[:3], lat[:3])



import numpy as np
import stripy

from scipy import spatial
from time import clock

try: range = xrange
except: pass


def test_area(mesh):

    def compute_area(v1, v2, v3):
        """ manually compute area on unit sphere """
        u1 = np.cross(v1, v2)
        u2 = np.cross(v2, v3)
        u3 = np.cross(v3, v1)

        s1 = np.sqrt(u1.dot(u1))
        s2 = np.sqrt(u2.dot(u2))
        s3 = np.sqrt(u3.dot(u3))

        u1 /= s1
        u2 /= s2
        u3 /= s3

        ca1 = -u1.dot(u3)
        ca2 = -u2.dot(u1)
        ca3 = -u3.dot(u2)

        a1 = np.arccos(ca1)
        a2 = np.arccos(ca2)
        a3 = np.arccos(ca3)

        areas = a1 + a2 + a3 - np.arccos(-1.0)
        return areas

    n  = mesh.npoints
    nt = mesh.simplices.shape[0]

    area1 = np.empty(nt)

    t = clock()
    for i in range(0,nt):
        tr = mesh.simplices[i]
        xi = mesh.x[tr]
        yi = mesh.y[tr]
        zi = mesh.z[tr]

        area1[i] = compute_area(xi, yi, zi)
    t1 = clock() - t
    t = clock()
    area2 = mesh.areas()
    t2 = clock() - t

    res = ((area1 - area2)**2).max()
    print("squared error in area calculation = {}\n  \
     - numpy took {}\n  \
     - stripy took {}".format(res, t1, t2))

    

def test_derivative(mesh):
    from scipy import interpolate

    # Create a field to test derivatives
    lons, lats = mesh.lons, mesh.lats + np.pi/2
    Z = np.exp(-lons**2 - lats**2)
    Zlons = -2*lons*Z
    Zlats = -2*lats*Z
    gradZ = np.hypot(Zlons, Zlats)

    # Stripy
    t = clock()
    Zx, Zy, Zz = mesh.gradient(Z, nit=10, tol=1e-10)
    t1 = clock() - t
    Zlons1, Zlats1 = mesh.transform_to_spherical(Zx, Zy, Zz)
    gradZ1 = np.hypot(Zlons1, Zlats1)

    # Spline
    spl = interpolate.SmoothSphereBivariateSpline(lats, lons, Z, s=1)
    t = clock()
    Zlons2 = spl.ev(lats, lons, dtheta=1)
    Zlats2 = spl.ev(lats, lons, dphi=1)
    t2 = clock() - t
    gradZ2 = np.hypot(Zlons2, Zlats2)

    res1 = ((gradZ1 - gradZ)**2).max()
    res2 = ((gradZ2 - gradZ)**2).max()
    print("squared error in first derivative\n  \
     - stripy = {} took {}s\n  \
     - spline = {} took {}s".format(res1, t1, res2, t2))


def test_interpolation(mesh):
    from scipy import interpolate

    lons, lats = mesh.lons, mesh.lats
    Z = np.exp(-lons**2 - lats**2)
    lon = 2.*np.pi*np.random.random(10)
    lat = np.arccos(2.*np.random.random(10) - 1.) - np.pi/2

    # Stripy
    zn1 = mesh.interpolate_nearest(lon, lat, Z)
    zl1 = mesh.interpolate_linear(lon, lat, Z)
    zc1 = mesh.interpolate_cubic(lon, lat, Z)

    # cKDTree
    tree = interpolate.NearestNDInterpolator((lons,lats), Z)
    zn2 = tree(lon, lat)

    # Least-squares spline
    tri = interpolate.LinearNDInterpolator((lons,lats), Z)
    zl2 = tri((lon, lat))


    # Cubic spline
    spl = interpolate.SmoothSphereBivariateSpline(lats+np.pi/2, lons, Z, s=1)
    zc2 = spl.ev(lat+np.pi/2,lon)

    print("squared residual in interpolation\n  \
     - nearest neighbour = {}\n  \
     - linear = {}\n  \
     - cubic  = {}".format(((zn1 - zn2)**2).max(), \
                           ((zl1 - zl2)**2).max(), \
                           ((zc1 - zc2)**2).max()))

def test_smoothing(mesh):
    pass

if __name__ == "__main__":
    
    # Create some (semi) random points
    np.random.seed(0)
    lons = 2.*np.pi*np.random.random(500)
    lats = np.arccos(2.*np.random.random(500) - 1.)

    # triangulation
    mesh = stripy.sTriangulation(lons, lats - np.pi/2)
    test_area(mesh)
    test_derivative(mesh)
    test_interpolation(mesh)



#!/usr/bin/python
# -*- coding: utf-8 -*-
import _stripack
import _ssrfpack
import numpy as np

try: range = xrange
except: pass

__version__ = "1.1"

_ier_codes = {0:  "no errors were encountered.",
              -1: "N < 3 on input.",
              -2: "the first three nodes are collinear.",
              -3: "duplicate nodes were encountered.",
              -4: "an error flag was returned by a call to SWAP in ADDNOD.\n \
                   This is an internal error and should be reported to the programmer.",
              'L':"nodes L and M coincide for some M > L.\n \
                   The linked list represents a triangulation of nodes 1 to M-1 in this case.",
              1: "NCC, N, NROW, or an LCC entry is outside its valid range on input.",
              2: "the triangulation data structure (LIST,LPTR,LEND) is invalid.",
              'K': 'NPTS(K) is not a valid index in the range 1 to N.'}


class sTriangulation(object):
    """
    Define a Delaunay triangulation for given points on a sphere
    where lons and lats are 1D numpy arrays of equal length.

    Algorithm
    ---------
     R. J. Renka (1997), Algorithm 772; STRIPACK: Delaunay triangulation
     and Voronoi diagram on the surface of a sphere"
     ACM Trans. Math. Softw., Vol. 23, No. 3, pp 416-434
     doi:10.1145/275323.275329

    Parameters
    ----------
     lons : 1D array of longitudinal coordinates in radians
     lats : 1D array of latitudinal coordinates in radians

    Attributes
    ----------
     lons : array of floats, shape (n,)
        stored longitudinal coordinates on a sphere
     lats : array of floats, shape (n,)
        stored latitudinal coordinates on a sphere
     x : array of floats, shape (n,)
        stored Cartesian x coordinates from input
     y : array of floats, shape (n,)
        stored Cartesian y coordinates from input
     z : array of floats, shape (n,)
        stored Cartesian z coordinates from input
     simplices : array of ints, shape (nsimplex, 3)
        indices of the points forming the simplices in the triangulation
        points are ordered anticlockwise
     lst : array of ints, shape (6n-12,)
        nodal indices with lptr and lend, define the triangulation as a set of N
        adjacency lists; counterclockwise-ordered sequences of neighboring nodes
        such that the first and last neighbors of a boundary node are boundary
        nodes (the first neighbor of an interior node is arbitrary).  In order to
        distinguish between interior and boundary nodes, the last neighbor of
        each boundary node is represented by the negative of its index.
        The indices are 1-based (as in Fortran), not zero based (as in python).
     lptr : array of ints, shape (6n-12),)
        set of pointers in one-to-one correspondence with the elements of lst.
        lst(lptr(i)) indexes the node which follows lst(i) in cyclical
        counterclockwise order (the first neighbor follows the last neighbor).
        The indices are 1-based (as in Fortran), not zero based (as in python).
     lend : array of ints, shape (n,)
        N pointers to adjacency lists.
        lend(k) points to the last neighbor of node K. 
        lst(lend(K)) < 0 if and only if K is a boundary node.
        The indices are 1-based (as in Fortran), not zero based (as in python).

    Notes
    -----
     Provided the nodes are randomly ordered, the algorithm
     has expected time complexity O(N*log(N)) for most nodal
     distributions.  Note, however, that the complexity may be
     as high as O(N**2) if, for example, the nodes are ordered
     on increasing latitude.
    """
    def __init__(self, lons, lats):

        # lons, lats = self._check_integrity(lons, lats)

        npoints = lons.size

        # compute cartesian coords on unit sphere.

        x, y, z = _stripack.trans(lats, lons)

        lst, lptr, lend, ierr = _stripack.trmesh(x, y, z)

        if ierr != 0:
            raise ValueError('ierr={} in trmesh\n{}'.format(ierr, _ier_codes[ierr]))

        self.lons = lons
        self.lats = lats
        self.npoints = npoints
        self.x = x
        self.y = y
        self.z = z
        self.lst = lst
        self.lptr = lptr
        self.lend = lend

        # Convert a triangulation to a triangle list form (human readable)
        # Uses an optimised version of trlist that returns triangles
        # without neighbours or arc indices
        nt, ltri, ierr = _stripack.trlist2(lst, lptr, lend)

        if ierr != 0:
            raise ValueError('ierr={} in trlist2\n{}'.format(ierr, _ier_codes[ierr]))

        # extract triangle list and convert to zero-based ordering
        self.simplices = ltri.T[:nt] - 1


    def _check_integrity(self, lons, lats):
        """
        Ensure lons and lats are:
         - 1D numpy arrays
         - equal size
         - within the appropriate range in radians
        """
        lons = np.array(lons).ravel()
        lats = np.array(lats).ravel()

        if len(lons.shape) != 1 or len(lats.shape) != 1:
            raise ValueError('lons and lats must be 1D')
        if lats.size != lons.size:
            raise ValueError('lons and lats must have same length')
        if (np.abs(lons)).max() > 2.*np.pi:
            raise ValueError("lons must be in radians (-2*pi <= lon <= 2*pi)")
        if (np.abs(lats)).max() > 0.5*np.pi:
            raise ValueError("lats must be in radians (-pi/2 <= lat <= pi/2)")
        return lons, lats


    def interpolate(self, lons, lats, zdata, order=1):
        """
        Base class to handle nearest neighbour, linear, and cubic interpolation.
        Given a triangulation of a set of nodes on the unit sphere, along with data
        values at the nodes, this method interpolates (or extrapolates) the value
        at a given longitude and latitude.

        Parameters
        ----------
         lons : float / array of floats, shape (l,)
            longitudinal coordinate(s) on the sphere
         lats : float / array of floats, shape (l,)
            latitudinal coordinate(s) on the sphere
         zdata : array of floats, shape (n,)
            value at each point in the triangulation
            must be the same size of the mesh
         order : int (default=1)
            order of the interpolatory function used
             0 = nearest-neighbour
             1 = linear
             3 = cubic
        
        Returns
        -------
         zi : float / array of floats, shape (l,)
            interpolates value(s) at (lons, lats)
        """
        lons, lats = self._check_integrity(lons, lats)

        if order not in [0,1,3]:
            raise ValueError("order must be 0, 1, or 3")
        if zdata.size != self.npoints:
            raise ValueError("data must be of size {}".format(self.npoints))

        zi, ierr = _stripack.interp_n(order, lats, lons,\
                                      self.x, self.y, self.z, zdata,\
                                      self.lst, self.lptr, self.lend)

        if ierr != 0:
            raise ValueError('ierr={} in interp_n\n{}'.format(ierr, _ier_codes[ierr]))

        return zi


    def interpolate_nearest(self, lons, lats, data):
        """
        Interpolate using nearest-neighbour approximation
        Returns the same as interpolate(lons,lats,data,order=0)
        """
        return self.interpolate(lons, lats, data, order=0)

    def interpolate_linear(self, lons, lats, data):
        """
        Interpolate using nearest-neighbour approximation
        Returns the same as interpolate(lons,lats,data,order=1)
        """
        return self.interpolate(lons, lats, data, order=1)

    def interpolate_cubic(self, lons, lats, data):
        """
        Interpolate using nearest-neighbour approximation
        Returns the same as interpolate(lons,lats,data,order=3)
        """
        return self.interpolate(lons, lats, data, order=3)


    def nearest_neighbour(self, lon, lat):
        """
        Get the index of the nearest neighbour to a point (lon,lat)
        and return the squared distance between (lon,lat) and
        its nearest neighbour.

        Notes
        -----
         Faster searches can be obtained using a KDTree.
         Store all x and y coordinates in a KDTree, then query
         a set of points to find their nearest neighbours.
        """
        # translate to unit sphere
        xi, yi, zi = _stripack.trans(lat, lon)

        # i is the node at which we start the search
        # the closest x coordinate is a good place
        i = ((self.x - xi)**2).argmin() + 1
        idx, d = _tripack.nearnd((xi,yi,zi), self.x, self.y, self.z, self.lst, self.lptr, self.lend, i)
        return idx - 1, d


    def gradient(self, f, nit=3, tol=1e-3, guarantee_convergence=False):
        """
        Return the gradient of an array.

        The method consists of minimizing a quadratic functional Q(G) over
        gradient vectors, where Q is an approximation to the linearized
        curvature over the triangulation of a C-1 bivariate function F(x,y)
        which interpolates the nodal values and gradients.

        Parameters
        ----------
         f : array of floats, shape (n,)
            field over which to evaluate the gradient
         nit: int (default: 3)
            number of iterations to reach a convergence tolerance, tol
            nit >= 1
         tol: float (default: 1e-3)
            maximum change in gradient between iterations.
            convergence is reached when this condition is met.

        Returns
        -------
         dfdx : array of floats, shape (n,)
            derivative of f in the x direction
         dfdy : array of floats, shape (n,)
            derivative of f in the y direction
         dfdz : array of floats, shape (n,)
            derivative of f in the z direction

        Notes
        -----
         For SIGMA = 0, optimal efficiency was achieved in testing with
         tol = 0, and nit = 3 or 4.

         The restriction of F to an arc of the triangulation is taken to be
         the Hermite interpolatory tension spline defined by the data values
         and tangential gradient components at the endpoints of the arc, and
         Q is the sum over the triangulation arcs, excluding interior
         constraint arcs, of the linearized curvatures of F along the arcs --
         the integrals over the arcs of D2F(T)**2, where D2F(T) is the second
         derivative of F with respect to distance T along the arc.
        """
        if f.size != self.npoints:
            raise ValueError('f should be the same size as mesh')

        gradient = np.zeros((3,self.npoints), order='F', dtype=np.float32)
        sigma = 0
        iflgs = 0

        ierr = 1
        while ierr == 1:
            ierr = _ssrfpack.gradg(self.x, self.y, self.z, f, self.lst, self.lptr, self.lend,\
                                   iflgs, sigma, nit, tol, gradient)
            if not guarantee_convergence:
                break

        if ierr < 0:
            raise ValueError('ierr={} in gradg\n{}'.format(ierr, _ier_codes[ierr]))

        return gradient[0], gradient[1], gradient[2]


    def transform_to_spherical(self, dfdx, dfdy, dfdz):
        """
        Transform the derivatives of f in the x,y,z directions into spherical
        derivatives.

        Arguments
        ---------
         dfdx : array of floats, shape (n,)
            derivative of f in the x direction
         dfdy : array of floats, shape (n,)
            derivative of f in the y direction
         dfdz : array of floats, shape (n,)
            derivative of f in the z direction

        Returns
        -------
         dfdlons : array of floats, shape (n,)
            derivatives of f w.r.t. longitude
         dfdlats : array of floats, shape (n,)
            derivatives of f w.r.t. latitude

        """
        cos_lons = np.cos(self.lons)
        sin_lons = np.sin(self.lons)

        cos_lats = np.cos(self.lats)
        sin_lats = np.sin(self.lats)

        dxdlons = -sin_lons*sin_lats
        dxdlats =  cos_lons*cos_lats

        dydlons =  cos_lons*sin_lats
        dydlats =  sin_lons*cos_lats

        dzdlons = -sin_lons
        dzdlats = 0.0

        # chain rule
        dfdlons = dfdx*dxdlons + dfdy*dydlons + dfdz*dzdlons
        dfdlats = dfdx*dxdlats + dfdy*dydlats

        return dfdlons, dfdlats


    def smoothing(self, f, w, sm, smtol, gstol):
        """
        Smooths a surface f by choosing nodal function values and gradients to
        minimize the linearized curvature of F subject to a bound on the
        deviation from the data values. This is more appropriate than interpolation
        when significant errors are present in the data.

        Parameters
        ----------
         f : array of floats, shape (n,)
            field to apply smoothing on
         w : array of floats, shape (n,)
            weights associated with data value in f
            w[i] = 1/sigma_f^2 is a good rule of thumb.
         sm : float
            positive parameter specifying an upper bound on Q2(f).
            generally n-sqrt(2n) <= sm <= n+sqrt(2n)
         smtol : float
            specifies relative error in satisfying the constraint
            sm(1-smtol) <= Q2 <= sm(1+smtol) between 0 and 1.
         gstol : float
            tolerance for convergence.
            gstol = 0.05*mean(sigma_f)^2 is a good rule of thumb.

        Returns
        -------
         f_smooth : array of floats, shape (n,)
            smoothed version of f
         (dfdx, dfdy, dfdz) : tuple of floats, tuple of 3 shape (n,) arrays
            first derivatives of f_smooth in the x, y, and z directions
        """
        if f.size != self.npoints or f.size != w.size:
            raise ValueError('f and w should be the same size as mesh')

        sigma = 0
        iflgs = 0
        prnt = -1

        f_smooth, df, ierr = _srfpack.smsurf(self.x, self.y, self.z, f, self.lst, self.lptr, self.lend,\
                                             iflgs, sigma, w, sm, smtol, gstol, prnt)

        if ierr < 0:
            raise ValueError('ierr={} in gradg\n{}'.format(ierr, _ier_codes[ierr]))
        if ierr == 1:
            raise RuntimeWarning("No errors were encountered but the constraint is not active --\n\
                  F, FX, and FY are the values and partials of a linear function \
                  which minimizes Q2(F), and Q1 = 0.")
        if ierr == 2:
            raise RuntimeWarning("if the constraint could not be satisfied to within SMTOL\
                  due to ill-conditioned linear systems.")

        return f_smooth, (df[0], df[1], df[2])


    def tri_area(self, lons, lats):
        """
        Calculate the area enclosed by 3 points on the unit sphere.

        Parameters
        ----------
         lons : array of floats, shape (3)
            longitudinal coordinates in radians
         lats : array of floats, shape (3)
            latitudinal coordinates in radians

        Returns
        -------
         area : float
            area of triangle on the unit sphere

        """
        lons, lats = self._check_integrity(lons, lats)

        # translate to unit sphere
        x, y, z = _stripack.trans(lats, lons)

        # compute area
        area = _stripack.areas(x, y, z)

        return area


    def areas(self):
        """
        Compute the area each triangle within the triangulation of points
        on the unit sphere.

        Returns
        -------
         area : array of floats, shape (nt,)
            area of each triangle in self.simplices where nt
            is the number of triangles.

        Notes
        -----
         This uses a Fortran 90 subroutine that wraps the AREA function
         to iterate over many points.
        """
        return _stripack.triareas(self.x, self.y, self.z, self.simplices.T+1)


import numpy as np
import stripy

from scipy import spatial
from time import clock

try: range = xrange
except: pass
    

def test_derivative(mesh):
    from scipy import interpolate

    # Create a field to test derivatives
    x, y = mesh.x, mesh.y
    Z = np.exp(-x**2 - y**2)
    Zx = -2*x*Z
    Zy = -2*y*Z
    gradZ = np.hypot(Zx, Zy)

    # Stripy
    t = clock()
    Zx1, Zy1 = mesh.gradient(Z, nit=10, tol=1e-10)
    t1 = clock() - t
    gradZ1 = np.hypot(Zx1, Zy1)

    # Spline
    spl = interpolate.SmoothBivariateSpline(x, y, Z)
    t = clock()
    Zx2 = spl.ev(x, y, dx=1)
    Zy2 = spl.ev(x, y, dy=1)
    t2 = clock() - t
    gradZ2 = np.hypot(Zx2, Zy2)

    # Clough Tocher
    # This one is most similar to what is used in stripy
    t = clock()
    cti = interpolate.CloughTocher2DInterpolator(np.column_stack([x,y]),\
                                                 Z, tol=1e-10, maxiter=20)
    t3 = clock() - t
    Zx3 = cti.grad[:,:,0].ravel()
    Zy3 = cti.grad[:,:,1].ravel()
    gradZ3 = np.hypot(Zx3, Zy3)

    res1 = ((gradZ1 - gradZ)**2).max()
    res2 = ((gradZ2 - gradZ)**2).max()
    res3 = ((gradZ3 - gradZ)**2).max()
    print("squared error in first derivative\n  \
           - stripy = {} took {}s\n  \
           - spline = {} took {}s\n  \
           - cloughtocher = {} took {}s".format(res1, t1, res2, t2, res3, t3))


def test_interpolation(mesh):
    from scipy import interpolate

    x, y = mesh.x, mesh.y
    Z = np.exp(-x**2 - y**2)
    # We ensure interpolation points are within convex hull
    xi = np.random.uniform(0.1, 0.9, 10)
    yi = np.random.uniform(0.1, 0.9, 10)

    # Stripy
    zn1 = mesh.interpolate_nearest(xi, yi, Z)
    zl1 = mesh.interpolate_linear(xi, yi, Z)
    zc1 = mesh.interpolate_cubic(xi, yi, Z)

    # cKDTree
    tree = interpolate.NearestNDInterpolator((x,y), Z)
    zn2 = tree(xi, yi)

    # Qhull
    tri = interpolate.LinearNDInterpolator((x,y), Z, 0.0)
    zl2 = tri((xi, yi))

    # Clough Tocher
    cti = interpolate.CloughTocher2DInterpolator(np.column_stack([x,y]),\
                                                 Z, tol=1e-10, maxiter=20)
    zc2 = cti((xi, yi))
    zc2[np.isnan(zc2)] = 0.0

    # Spline
    spl = interpolate.SmoothBivariateSpline(x, y, Z)
    zc3 = spl.ev(xi,yi)

    # Radial basis function
    rbf = interpolate.Rbf(x, y, Z)
    zc4 = rbf(xi, yi)

    print("squared residual in interpolation\n  \
           - nearest neighbour = {}\n  \
           - linear = {}\n  \
           - cubic (clough-tocher) = {}\n  \
           - cubic (spline) = {}\n  \
           - cubic (rbf) = {}".format(((zn1 - zn2)**2).max(), \
                                      ((zl1 - zl2)**2).max(), \
                                      ((zc1 - zc2)**2).max(), \
                                      ((zc1 - zc3)**2).max(), \
                                      ((zc1 - zc4)**2).max(),) )


def test_smoothing(mesh):
    pass

if __name__ == "__main__":
    
    # Create some (semi) random points
    np.random.seed(0)
    x = np.random.random(500)
    y = np.random.random(500)

    # triangulation
    mesh = stripy.Triangulation(x, y)

    test_derivative(mesh)
    test_interpolation(mesh)


from numpy.distutils.core import setup, Extension

# interface for Renka's algorithm 772 fortran code
ext1 = Extension(name    = '_stripack',
                 sources = ['src/stripack.pyf','src/stripack.f90'])
ext2 = Extension(name    = '_tripack',
                 sources = ['src/tripack.pyf', 'src/tripack.f90'])
ext3 = Extension(name    = '_srfpack',
                 sources = ['src/srfpack.pyf', 'src/srfpack.f'])
ext4 = Extension(name    = '_ssrfpack',
                 sources = ['src/ssrfpack.pyf', 'src/ssrfpack.f'])

if __name__ == "__main__":
    setup(name = 'stripy',
          author            = "LM",
          author_email      = "louis.moresi@unimelb.edu.au",
          url               = "https://github.com/University-of-Melbourne-Geodynamics/stripy",
          version           = "0.1.1",
          description       = "Python interface to TRIPACK and STRIPACK fortran code for triangulation/interpolation in Cartesian coordinates and on a sphere",
          ext_modules       = [ext1, ext2, ext3, ext4],
          packages          = ['stripy'],
          classifiers       = ['Programming Language :: Python :: 2',
                               'Programming Language :: Python :: 2.6',
                               'Programming Language :: Python :: 2.7',
                               'Programming Language :: Python :: 3',
                               'Programming Language :: Python :: 3.3',
                               'Programming Language :: Python :: 3.4',
                               'Programming Language :: Python :: 3.5',
                               'Programming Language :: Python :: 3.6']
          )
